#ifndef __LINEAR_REGRESSION_H__
#define __LINEAR_REGRESSION_H__
/*********************************************************************
 * Software License Agreement (Modified BSD License)
 *
 * Copyright (c) 2010, Daniel Munoz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the copyright holders' organizations nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#include <string>
#include <map>
#include <vector>
#include <iostream>
#include <fstream>

#include <m3n/regressors/regressor_wrapper.h>

#include <Eigen/Core>
#include <Eigen/LU>

// --------------------------------------------------------------
/*!
 * \brief Implements simple linear regression
 */
// --------------------------------------------------------------
class LinearRegression: public RegressorWrapper
{
  public:
    // --------------------------------------------------------------
    /**
     * \brief Instantiates linear regression with no regularization
     */
    // --------------------------------------------------------------
    LinearRegression();

    // --------------------------------------------------------------
    /**
     * \brief Instantiates linear regression with specified regularization
     *
     * \param regularization Tikhonov parameter
     *
     * \warning This LinearRegression is forever bound to these parameters
     */
    // --------------------------------------------------------------
    LinearRegression(double regularization);

    // --------------------------------------------------------------
    /**
     * \brief See RegressorWrapper::paramClone()
     */
    // --------------------------------------------------------------
    virtual RegressorWrapper* paramClone() const;

    // --------------------------------------------------------------
    /**
     * \brief See RegressorWrapper::saveToFile()
     */
    // --------------------------------------------------------------
    virtual int saveToFile(const std::string directory,
                           const std::string basename);

    // --------------------------------------------------------------
    /**
     * \brief See RegressorWrapper::loadFromFile()
     */
    // --------------------------------------------------------------
    virtual int loadFromFile(const std::string directory,
                             const std::string basename);

  protected:
    // --------------------------------------------------------------
    /**
     * \brief See RegressorWrapper::doClear()
     */
    // --------------------------------------------------------------
    virtual void doClear();

    // --------------------------------------------------------------
    /**
     * \brief See RegressorWrapper::doTrain()
     */
    // --------------------------------------------------------------
    virtual int doTrain(const std::vector<const std::vector<double>*>& interm_feature_vals,
                        const std::vector<unsigned int>& interm_start_idx,
                        const std::vector<double>& interm_target);

    // --------------------------------------------------------------
    /**
     * \brief See RegressorWrapper::doPredict()
     */
    // --------------------------------------------------------------
    virtual int doPredict(const std::vector<double>& feature_vals,
                          const unsigned int start_idx,
                          double& predicted_val) const;

  private:
    /*! \brief Tikhonov regularization parameter */
    double m_regularization;
    std::vector<double> m_weights;
};

#endif
